%--------------------------------------------------------------------------
%
% This program demonstrates the main ideas described in Kanan and Cottrell
% (2010). I reimplemented it, so the results might be a little different
% than the paper. I have also encapsulated the saliency map and feature
% extraction code in Run_SUN.m.
%
% I've included a portion of the butterflies dataset by Lazebnik, Schmid, 
% & Ponce (2004), but I've deleted some of the images and resized them to
% save storage. I wanted to ensure that the code could be run immediately.
% The model works very well on the AR face dataset, and I would have 
% preferred to include the first 10 people from that dataset, but it is not
% redistributable. The full butterflies dataset is available here: 
% http://www-cvr.ai.uiuc.edu/ponce_grp/data/
%
% If you use any of the code, please cite my paper. 
%
%
% Author:   Christopher Kanan (chriskanan@gmail.com)
%
% Cite as:  Kanan, C. & Cottrell, G. W. (2010). Robust classification of
%             objects, faces, and flowers using natural image Statistics.
%             In Proceedings of the IEEE Conference on Computer Vision
%             and Pattern Recognition (CVPR), 2010.
%==========================================================================


function varargout = Demo(varargin)
% Begin initialization code - DO NOT EDIT
gui_Singleton = 1;
gui_State = struct('gui_Name',       mfilename, ...
    'gui_Singleton',  gui_Singleton, ...
    'gui_OpeningFcn', @Demo_OpeningFcn, ...
    'gui_OutputFcn',  @Demo_OutputFcn, ...
    'gui_LayoutFcn',  [] , ...
    'gui_Callback',   []);
if nargin && ischar(varargin{1})
    gui_State.gui_Callback = str2func(varargin{1});
end

if nargout
    [varargout{1:nargout}] = gui_mainfcn(gui_State, varargin{:});
else
    gui_mainfcn(gui_State, varargin{:});
end
% End initialization code - DO NOT EDIT

%--------------------------------------------------------------------------
% Executes just before Demo is made visible.
%--------------------------------------------------------------------------
function Demo_OpeningFcn(hObject, eventdata, handles, varargin)
% This function has no output args, see OutputFcn.
% hObject    handle to figure
% eventdata  reserved - to be defined in a future version of MATLAB
% handles    structure with handles and user data (see GUIDATA)
% varargin   command line arguments to Demo (see VARARGIN)

% Choose default command line output for Demo
%--------------------------------------------------------------------------
handles.output = hObject;

%various parameters
handles.Dataset_Index = 1;
handles.animate = true;
handles.Model.SUN = [];
handles.retina_radius = 25;
handles.max_train_samples = 100;
handles.max_test_samples = 100;
handles.max_principal_components = 500;


handles.Dataset_List = {'Butterflies Dataset' }; 



axes_list = [handles.axes_image, handles.axes_performance, handles.axes_saliency];
for k = 1:length(axes_list)
    set(axes_list, 'XTickLabel', [], 'YTickLabel', [], 'XTick', [], 'YTick', []);
end



% Update handles structure
guidata(hObject, handles);

% UIWAIT makes Demo wait for user response (see UIRESUME)
% uiwait(handles.figure1);

%--------------------------------------------------------------------------
% Outputs from this function are returned to the command line.
%--------------------------------------------------------------------------
function varargout = Demo_OutputFcn(hObject, eventdata, handles)
% varargout  cell array for returning output args (see VARARGOUT);
% hObject    handle to figure
% eventdata  reserved - to be defined in a future version of MATLAB
% handles    structure with handles and user data (see GUIDATA)

% Get default command line output from handles structure
varargout{1} = handles.output;


%--------------------------------------------------------------------------
% Executes on button press in btnTrain. This function acquires samples
% which are used to train the model with.
%--------------------------------------------------------------------------
function btnTrain_Callback(hObject, eventdata, handles)
% hObject    handle to btnTrain (see GCBO)
% eventdata  reserved - to be defined in a future version of MATLAB
% handles    structure with handles and user data (see GUIDATA)


%num_train = 1;
%num_test = inf;
%max_class = inf;
max_class = inf;

num_train = 5;
num_test = 5;



handles.max_filters = inf;

dataset_name = 'Butterflies Dataset';

[handles.train_files, handles.test_files, handles.Class_Names] = Load_Demo_Dataset(dataset_name, max_class, num_train, num_test);
%[handles.train_files, handles.test_files, handles.Class_Names] = Load_Dataset(dataset_name, max_class, num_train, num_test);
handles.test_files = handles.test_files(:, 1:min(num_test, size(handles.test_files, 2)));

handles = Train_And_Test_Model(handles, 1);

% Update handles structure
guidata(hObject, handles);


%this will make it test without clicking
%handles = Train_And_Test_Model(handles, 2);


%--------------------------------------------------------------------------
% Executes on button press in btnTest.
%--------------------------------------------------------------------------
function btnTest_Callback(hObject, eventdata, handles)
% hObject    handle to btnTest (see GCBO)
% eventdata  reserved - to be defined in a future version of MATLAB
% handles    structure with handles and user data (see GUIDATA)

handles = Train_And_Test_Model(handles, 2);


%--------------------------------------------------------------------------
% Parses training/testing images
%--------------------------------------------------------------------------
function [handles, Stats] = Train_And_Test_Model(handles, operating_mode)

TRAIN_MODE = 1;
TEST_MODE = 2;

handles.Permute_Mat = [];
handles.grid_list = [1, 2, 4]; %for spatial pooling
img_size = 128;
%figure out how long each vector is in a dumb way (but it makes it easier
%to tinker with the code)
[Saliency_Map, Feature_Maps, ICA_Maps, img, handles.Model.SUN] = Run_SUN(rand(128, 128, 3, 'single'), handles.Model.SUN, handles.max_filters);
Saliency_Map = [];
ICA_Maps = [];
img = [];
feat = Fast_Spatial_Pooling(zeros(handles.retina_radius, handles.retina_radius, size(Feature_Maps, 3)), handles.grid_list, []);
Feature_Maps = [];

feat_length = numel(feat);

set(handles.txt_accuracy, 'String', '');


if operating_mode == TRAIN_MODE
    handles.Net = [];
    handles.PCA_Data = [];
    files = handles.train_files;
    max_samples = numel(files)*handles.max_train_samples; %determine the maximum number
    Training_Data = zeros(feat_length, max_samples, 'single');
    Training_Location = zeros(2, max_samples, 'single');
    Training_Labels = zeros(1, max_samples, 'single');
      
    num_fixations = handles.max_train_samples;
    Display_Message(handles, 'Acquiring training samples...');
    handles.Perf_Img = [];
    
elseif operating_mode == TEST_MODE
    files = handles.test_files;
    num_fixations = handles.max_test_samples;
    Display_Message(handles, 'Evaluating model...');
else
    error('Unknown operating_mode.');
end

Update_Accuracy(handles, '');


sample_start = 1;

online_total = 0;
online_correct = 0;
class_correct = zeros(1, size(files, 1));
class_total = zeros(1, size(files, 1));

rand('state', 0);


temporal_perf = 0;

total = 0;
for class = 1:size(files, 1)
    for instance = 1:size(files, 2)
        fname = files{class, instance};
        img = [];
        if ~isempty(fname)
            try
                img = imread(fname);
            end
        end
        
        if ~isempty(img)
            if operating_mode == TEST_MODE % clear display image for performance                
                Pretty_Display(handles.axes_performance, handles.Perf_Img, false, true);
            end
            
            %resize the image
            [sy, sx, sz] = size(img);
            rescale_factor = img_size / min(sy, sx);
            img = single(img);
            if max(img(:)) > 1, img = img ./ 255;  end
            img = imresize(img, round([sy, sx]*rescale_factor));
            img = max(min(img, 1), 0);
                                    
            Pretty_Display(handles.axes_image, img); %show image
            Pretty_Display(handles.axes_saliency, zeros(size(img), 'single')); %clear saliency map
            
            if operating_mode == TRAIN_MODE && instance == 1 %make display image for performance
                handles.Perf_Img = Prepare_Performance_Image(handles.Perf_Img, uint8(round(img*255)), class, size(files, 1));
                Pretty_Display(handles.axes_performance, handles.Perf_Img, false, true);
            end
            
            drawnow;
            %Get the Saliency Map and Features
            Display_Message(handles, 'Acquiring ICA maps...');            
            [Saliency_Map, Feature_Maps] = Run_SUN(img, handles.Model.SUN, handles.max_filters);
            Display_Message(handles, 'Sampling features...');
            
            ICA_Maps = [];
            img = [];
            Pretty_Display(handles.axes_saliency, Saliency_Map, true); %show saliency map
            drawnow;
            
            total = total + 1;
            %perform classification using fixations
            [handles, Samples, Locations, correct, temporal_k] = Process_Samples(handles, operating_mode, Saliency_Map, Feature_Maps, num_fixations, class);
            
            if operating_mode == TRAIN_MODE
                sample_stop = sample_start + (size(Samples, 2) - 1);
                Training_Data(:, sample_start:sample_stop) = Samples;
                Training_Labels(sample_start:sample_stop) = class;
                Training_Location(:, sample_start:sample_stop) = Locations;
                sample_start = sample_stop + 1;                       
                                
            else
                online_total = online_total + 1;
                class_total(class) = class_total(class) + 1;
                
                online_correct = online_correct + correct;               
                class_correct(class) = class_correct(class) + correct;
                
                Update_Accuracy(handles, sprintf('Accuracy: %.2f', 100*(online_correct / online_total)));
                temporal_perf = temporal_perf + temporal_k;
            end
            
            Samples = [];
            pause(0.5);
        end
        
        clear Saliency_Map Feature_Maps img
        %Saliency_Map = [];
        %Feature_Maps = [];
        %img = [];
    end
end

if operating_mode == TRAIN_MODE
    Training_Data = Training_Data(:, 1:sample_stop);
    Training_Location = Training_Location(:, 1:sample_stop);
    Training_Labels = Training_Labels(1:sample_stop);
end


if operating_mode == TRAIN_MODE
    %now do PCA
    Display_Message(handles, 'Performing PCA...');
    [handles.PCA_Data.Mean, handles.PCA_Data.Vectors, Training_Data] = Whitened_PCA(Training_Data, 500, false);
    
    handles.Net = NN_KDE_Train(Training_Data, Training_Labels);
    handles.Net.W_Locations = Training_Location;   
    Display_Message(handles, 'Ready for evaluation!');
    
else
    Display_Message(handles, 'Testing complete!');    
    temporal_perf = temporal_perf ./ total;
    
    %figure;plot(temporal_perf)    
end




%--------------------------------------------------------------------------
% Perform fixations and classify them
%--------------------------------------------------------------------------
function [handles, Samples, Locations, correct, temporal_perf] = Process_Samples(handles, operating_mode, Saliency_Map, Feature_Maps, num_fixations, class)

sorted_map = [];
Samples = [];
Permute_Mat = [];
correct = [];
reduce_dimensionality = false;
if isfield(handles, 'PCA_Data') && ~isempty(handles.PCA_Data)
    reduce_dimensionality = true;
end

Locations = zeros(2, num_fixations, 'single');
Class_Prob = [];
Prior = [];

temporal_perf = zeros(1, num_fixations);
for t = 1:num_fixations        
    [new_center, Saliency_Map, sorted_map] = Saccade(Saliency_Map, handles.retina_radius, sorted_map);    
    x_ndx = new_center(2);
    y_ndx = new_center(1);
    
    %just in case we saccade off of the image. This *should* not be
    %necessary unless you tinker with Saccade
    if x_ndx <= handles.retina_radius
        x_ndx = handles.retina_radius+1;
    elseif x_ndx > size(Saliency_Map, 2) - handles.retina_radius
        x_ndx = size(Saliency_Map, 2) - handles.retina_radius;
    end
    if y_ndx <= handles.retina_radius
        y_ndx = handles.retina_radius+1;
    elseif y_ndx > size(Saliency_Map, 1) - handles.retina_radius
        y_ndx = size(Saliency_Map, 1) - handles.retina_radius;
    end
    
    Locations(:, t) = [y_ndx;x_ndx] ./ [size(Saliency_Map, 1); size(Saliency_Map, 2)];
    if handles.animate
        handles = Draw_Retina(handles, y_ndx, x_ndx);
    end
    
    %get a square stack of ICA features
    Patch_Stack = Feature_Maps(y_ndx-handles.retina_radius:y_ndx+handles.retina_radius, x_ndx-handles.retina_radius:x_ndx+handles.retina_radius, :);
    
    %do spatial pooling
    [feat, Permute_Mat] = Fast_Spatial_Pooling(Patch_Stack, handles.grid_list, Permute_Mat);
    feat = feat(:);
    feat = feat ./ sqrt(feat'*feat); %make unit length
        
    if reduce_dimensionality %do PCA
        feat = handles.PCA_Data.Vectors*(feat - handles.PCA_Data.Mean);
    end
    
    if isempty(Samples)
        %allocate memory
        Samples = zeros(length(feat), num_fixations, 'single');
    end
    
    Samples(:, t) = feat;
    
    
    if operating_mode == 2 %test mode
        [Current_Prob, Class_Prob] = NIMBLE_Classify(handles.Net, feat, Locations(:, t), Class_Prob, Prior);
        [k_val, k_max] = max(Current_Prob);
        
        temporal_perf(t) = k_max == class;
        
        Display_Img = Update_Performance_Image(handles.Perf_Img, Current_Prob, class);
        Pretty_Display(handles.axes_performance, Display_Img, false, true);
        Display_Img = [];
        drawnow;
    end
end

if operating_mode == 2 %test mode
    [val, max_prob_ndx] = max(Current_Prob);
    correct = (class == max_prob_ndx);
end

%--------------------------------------------------------------------------
% Saccade to next target
%--------------------------------------------------------------------------
function [new_center, saliency_map, sorted_map] = Saccade(saliency_map, radius, sorted_map)

if isempty(saliency_map)
    error('Saliency map is empty!');
end

initial_map = saliency_map;

saliency_map = saliency_map(radius+1:end-radius, radius+1:end-radius);

if isempty(sorted_map) %sort map if necessary
    saliency_map = double(saliency_map);
    saliency_map = saliency_map ./ sum(saliency_map(:));
    [s_vals, s_indices] = sort(saliency_map(:), 1, 'descend');
    sorted_map.s_vals = cumsum(s_vals);
    sorted_map.s_indices = s_indices;
end

%make a fixation
s_vals = sorted_map.s_vals;
s_indices = sorted_map.s_indices;
r = rand(1);
ndx = s_indices(find(r <= s_vals, 1, 'first'));

[y_ndx, x_ndx] = ind2sub([size(saliency_map, 1), size(saliency_map, 2)], ndx);

y_ndx = y_ndx + radius;
x_ndx = x_ndx + radius;

new_center = [y_ndx, x_ndx];
saliency_map = initial_map;


%--------------------------------------------------------------------------
% Draw the retina
%--------------------------------------------------------------------------
function handles = Draw_Retina(handles, y_ndx, x_ndx)

r = handles.retina_radius;
w = 2*handles.retina_radius;

pos =  [x_ndx-r, y_ndx-r, w, w];

try
    set(handles.Retina.h_image, 'Position', pos);
    set(handles.Retina.h_saliency, 'Position', pos);
catch
    handles.Retina.h_image = rectangle('Position', pos, 'LineWidth', 3, 'EdgeColor', [0, 1, 0], 'Curvature', [0 0], 'Parent', handles.axes_image);
    handles.Retina.h_saliency = rectangle('Position', pos, 'LineWidth', 3, 'EdgeColor', [0, 1, 0], 'Curvature', [0 0], 'Parent', handles.axes_saliency);
end

drawnow;

%pause(.001); %amount of time to pause

%--------------------------------------------------------------------------
% Displays a message
%--------------------------------------------------------------------------
function Display_Message(handles, msg)

set(handles.txtMessage, 'String', msg);
drawnow;

%--------------------------------------------------------------------------
% Update accuracy
%--------------------------------------------------------------------------
function Update_Accuracy(handles, msg)

set(handles.txt_accuracy, 'String', msg);
drawnow;



%--------------------------------------------------------------------------
% Update performance image
%--------------------------------------------------------------------------
function Display_Img = Update_Performance_Image(Perf_Img, Current_Prob, class)

Display_Img = Perf_Img;

default_size = size(Perf_Img, 1) / 2;

filled = round(default_size * Current_Prob);
filled_start = max(default_size - filled, 1);
x_stop = 0;
for k = 1:length(Current_Prob)   
    if k == class
        color = [0, 1, 0];
    else
        color = [1, 0, 0];
    end
    
    x_stop = x_stop + default_size;
    x_start = 1 + x_stop - default_size;
    
    for c = 1:3
        Display_Img(filled_start(k):default_size, x_start:x_stop, c) = color(c);
    end
end


%--------------------------------------------------------------------------
% Prepare performance image
%--------------------------------------------------------------------------
function Perf_Img = Prepare_Performance_Image(Perf_Img, New_Img, class, max_class)

height = 256;

default_size = height / 2;

width = max_class * default_size;
if isempty(Perf_Img)
    Perf_Img = zeros(height, width, 3, 'single');    
end

New_Img = single(New_Img) / 255;
New_Img = imresize(New_Img, [default_size, default_size]);
New_Img = max(min(New_Img, 1), 0);

if size(New_Img, 3) == 1 %make sure B&W images still work
    New_Img = repmat(New_Img, [1, 1, 3]);
end

start = default_size * (class - 1) + 1;
Perf_Img(default_size+1:end, start:start+default_size-1, :) = New_Img;
