%=========================================================================
function [train_files, test_files, Class_Names] = Load_Demo_Dataset(dataset_name, max_class, train_num, test_num)

base_dir = [];

BB_Dir = fullfile(base_dir, 'Butterflies Dataset');

if strcmpi(dataset_name, 'Butterflies Dataset')
    [Files, File_Counts, Class_Names] = Load_Folder_Dataset(BB_Dir);
    chop_up_files = true;    
else
    error('Unknown dataset name.');
end

if chop_up_files %used for some datasets to turn them into training and test sets
    total_classes = max_class;
    max_test = min(test_num, min(max(File_Counts) - train_num));
    
    f_size = min(total_classes, size(Files, 1));
    
    train_files = cell(f_size, train_num);
    test_files = cell(f_size, max_test);
    
    for f = 1:f_size
        list = randperm(File_Counts(f));
        train_list = list(1:train_num);
        test_list = list(train_num+1:end);
        test_len = min(max_test, length(test_list));
        test_list = test_list(1:test_len);
        train_files(f, :) = Files(f, train_list);
        test_files(f, 1:length(test_list)) = Files(f, test_list);
    end
end

return

%=========================================================================
function [Files, File_Counts, Class_Names] = Load_Folder_Dataset(Base_DIR)


Files = cell(305, 2000);
Class_Names = cell(105, 1);
File_Counts = zeros(105, 1);
folders = dir(Base_DIR);

class_count = 0;
max_files = 0;
for k = 1:length(folders)
   folder_name = folders(k).name;
   if length(folder_name) <= 2
   else
       folder_files = dir(fullfile(Base_DIR, folder_name));
       class_count = class_count + 1;
       file_count = 0;
       
       Class_Names{class_count} = folder_name;
       
       f_list = randperm(length(folder_files)); %randomize!
       for f = f_list 
           file_name = folder_files(f).name;
           if length(file_name) > 4 && ~isempty(strfind(file_name, '.'))
                file_name = fullfile(Base_DIR, folder_name, file_name);
                file_count = file_count + 1;
                Files{class_count, file_count} = file_name;
           end
       end
       File_Counts(class_count) = file_count;
       max_files = max(file_count, max_files);
   end
end

File_Counts = File_Counts(1:class_count);
Files = Files(1:class_count, 1:max_files);
Class_Names = Class_Names(1:class_count);

%=========================================================================